# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import codecs
import sys
import io
import time
import datetime
import threading
import platform
from collections import namedtuple
# import multiprocessing
# sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="UTF-8")
# sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding="UTF-8")

import logging
from arjuna.core.utils import thread_utils
from arjuna.core.thread.decorators import *
from arjuna.core.utils import sys_utils
from arjuna.core.adv.proxy import ROProxy
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.tpi.error import UndefinedConfigError
from arjuna.core.adv.decorators import singleton
from arjuna.tpi.error import TestSelectorNotFoundError
import codecs
import sys

LinkedArjunaProject = namedtuple("LinkedArjunaProject", "name, location, ref_conf_editable, ref_conf, data_env_conf_map")

@singleton
class ArjunaSingleton:

    def __init__(self):
        self.__project_root_dir = None
        self.__test_session = None
        self.__config_map = dict()
        self.__run_id = None

        # From central config
        self.prop_enum_to_prop_path_map = {}
        # It is base config object
        self.arjuna_options = {}

        # Check relevance of above and following
        self.prog = "Arjuna"

        self.dl = None
        self.log_file_discovery_info = False
        self.__contextual_data_references = None
        self.__indexed_data_references = None
        self.__logger = None
        from arjuna.engine.data.store import DataStore
        self.__data_store = DataStore()
        self.__thread_wise_group_params_map = dict()
        self.__thread_wise_ref_conf_map = dict()
        self.__thread_wise_test_selector_map = dict()
        self.__thread_wise_group_command_map = dict()
        self.__test_meta_data = dict()

        from arjuna.tpi.engine.testwise import CurrentTestWiseContainer
        self.__current_test_wise_objects = CurrentTestWiseContainer()
        self.__allowed_log_contexts = {"default"}
        self.__bmproxy_server = None
        from arjuna.interact.gui.auto.finder.withx import WithX
        self.__common_withx_ref = WithX()

        self.__linked_projects = None

    def __start_bmproxy(self, config):
        # BrowserMob
        from browsermobproxy import Server
        from arjuna.tpi.constant import ArjunaOption
        capture_traffic = config.value(ArjunaOption.BROWSER_NETWORK_RECORDER_ENABLED)
        if capture_traffic:
            bmproxy_dir = config.value(ArjunaOption.TOOLS_BMPROXY_DIR)
            sub_dirs = os.listdir(bmproxy_dir)
            bmproxy_bin_dir = None
            if "bin" in sub_dirs:
                bmproxy_bin_dir = os.path.join(bmproxy_dir, "bin")
            else:
                sub_dirs.sort(reverse=True) # Last version will be picked.
                for d in sub_dir:
                    if d.startswith("browsermob"):
                        bmproxy_bin_dir = os.path.join(bmproxy_dir, d, "bin")
                        break

            if bmproxy_bin_dir is None:
                raise Exception("Network recording is enabled in configuration. There was an error in creating proxy server/server using BrowserMob Proxy. Could not find proxy package at {}".format(bmproxy_dir))
            
            if platform.system().lower() == "windows":
                exe = "browsermob-proxy.bat"
            else:
                exe = "browsermob-proxy"
            bmproxy_exe_path = os.path.join(bmproxy_bin_dir, exe)

            try:
                self.__bmproxy_server = Server(bmproxy_exe_path)
                self.__bmproxy_server.start()
            except ProxyServerError as e:
                raise Exception("Network recording is enabled in configuration. There was an error in creating proxy server/server using BrowserMob Proxy. Fix and retry. Error message: {}".format(str(e)))

    @property
    def bmproxy_server(self):
        return self.__bmproxy_server

    @property
    def allowed_log_contexts(self):
        return self.__allowed_log_contexts

    @property
    def gui_mgr(self):
        return self.__test_session.gui_manager

    def __create_dir_if_doesnot_exist(self, d):
        if not os.path.exists(d):
            os.makedirs(d)

    def init(self, project_root_dir, cli_config, run_id, *, static_rid, linked_projects):
        from arjuna.configure.options import ArjunaOptions
        ArjunaOptions.load_desc()
        
        self.__project_root_dir = project_root_dir

        from arjuna.engine.controller import TestSessionController
        self.__test_session = TestSessionController()
        run_id = run_id and run_id or "mrun"
        prefix = ""
        if not static_rid:
            prefix = "{}-".format(datetime.datetime.now().strftime("%Y.%m.%d-%H.%M.%S.%f")[:-3])
        run_id = "{}{}".format(prefix, run_id)

        # Process linked Arjuna projects
        def get_arjuna_project_path_and_name(fpath):
            from arjuna.core.utils import file_utils
            if not file_utils.is_absolute_path(fpath):
                fpath = os.path.abspath(os.path.join(project_root_dir, fpath))
            
            if not file_utils.is_dir(fpath):
                if file_utils.is_file(fpath):
                    raise Exception("The Linked Arjuna Project path is a file. It should be a directory: {}".format(fpath))
                else:
                    raise Exception("The Linked Arjuna Project path does not exist: {}".format(fpath))
            else:
                if not os.path.exists(os.path.join(fpath, "script", "arjuna_launcher.py")):
                    raise Exception("The Linked Arjuna Project path exists but is not an Arjuna test project: {}".format(fpath))
                return os.path.basename(fpath), fpath, os.path.abspath(fpath + "/..") 

        self.__linked_projects = list()
        linked_project_dict = dict()

        from arjuna.tpi.constant import ArjunaOption

        unique_paths = list()
        for arjuna_proj_dir in linked_projects:
            proj_name, proj_path, proj_import_path = get_arjuna_project_path_and_name(arjuna_proj_dir)
            from arjuna.configure.configurator import TestConfigurator
            l_proj_configurator = TestConfigurator(proj_path, cli_config, run_id)
            ref_conf_editable = l_proj_configurator.ref_config
            ref_conf = self.__test_session._create_config(ref_conf_editable)
            data_env_conf_map = l_proj_configurator.file_confs
            #for run_env_conf in [self.__test_session._create_config(econf, name=name) for name, econf in l_proj_configurator.file_confs.items()]:
            #    data_env_conf_map[run_env_conf.name] = run_env_conf
            self.__linked_projects.append(LinkedArjunaProject(name=proj_name, location=proj_path, ref_conf=ref_conf, ref_conf_editable=ref_conf_editable, data_env_conf_map=data_env_conf_map))
            unique_paths.append(proj_import_path)
        
        unique_paths = set(unique_paths)
        for p in unique_paths:
            sys.path.append(p)

        self.__thread_wise_ref_conf_map[threading.currentThread().name] = self.__test_session.init(project_root_dir, cli_config, run_id, self.__linked_projects)

        def get_src_file_path(src):
            return os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), src))

        def get_proj_target_path(dest):
            return os.path.join(self.ref_config.value(ArjunaOption.PROJECT_ROOT_DIR), dest)

        def copy_file(src, dest):
            shutil.copyfile(get_src_file_path(src), get_proj_target_path(dest))

        res_import_block = '''
try:
    from {project}.lib.resource import *
except ModuleNotFoundError as e:
    if e.name not in {{"{project}.lib", "{project}.lib.resource"}}:
        raise Exception(e.name)
'''

        from arjuna import Arjuna

        f = open(get_src_file_path("../../res/conftest.txt"), "r")
        contents = f.read()
        f.close()
        res_import_blocks = list()
        for proj in self.__linked_projects:
            res_import_blocks.append(res_import_block.format(project=proj.name))
        res_import_blocks.append(res_import_block.format(project=self.ref_config.value(ArjunaOption.PROJECT_NAME)))
        contents = contents.format(res_import_block="".join(res_import_blocks))
        f = open(get_proj_target_path("test/conftest.py"), "w")
        f.write(contents)
        f.close()

        self.__create_dir_if_doesnot_exist(self.ref_config.value(ArjunaOption.REPORT_DIR))
        self.__create_dir_if_doesnot_exist(self.ref_config.value(ArjunaOption.REPORT_XML_DIR))
        self.__create_dir_if_doesnot_exist(self.ref_config.value(ArjunaOption.REPORT_HTML_DIR))
        self.__create_dir_if_doesnot_exist(self.ref_config.value(ArjunaOption.LOG_DIR))
        self.__create_dir_if_doesnot_exist(self.ref_config.value(ArjunaOption.SCREENSHOTS_DIR))

        from arjuna.engine.logger import Logger
        self.__logger = Logger(self.ref_config)
        from arjuna import ArjunaOption
        self.__allowed_log_contexts = self.ref_config.value(ArjunaOption.LOG_ALLOWED_CONTEXTS)

        from arjuna.tpi.hook.config import Configurator
        configurator = Configurator()
        # Load configs from config hooks
        hooks_dir = self.ref_config.value(ArjunaOption.HOOKS_DIR)
        if os.path.isdir(hooks_dir):
            sys.path.append(hooks_dir)
        try:
            from arjuna_config import register_ref_confs
        except ModuleNotFoundError as e: # Module not defined.
            pass
        except ImportError as f: # Hook not defined
            pass
        else:
            register_ref_confs(configurator)

        def get_deps_dir_path(fpath):
            from arjuna.core.utils import file_utils
            if file_utils.is_absolute_path(fpath):
                if not file_utils.is_dir(fpath):
                    if file_utils.is_file(fpath):
                        raise Exception("Not a directory: {}".format(fpath))
                return fpath
            else:
                fpath = os.path.abspath(os.path.join(self.ref_config.value(ArjunaOption.PROJECT_ROOT_DIR), fpath))
                if not file_utils.is_dir(fpath):
                    if file_utils.is_file(fpath):
                        raise Exception("Not a directory: {}".format(fpath))
                return fpath

        deps_dir = get_deps_dir_path(self.ref_config.value(ArjunaOption.DEPS_DIR))
        if os.path.isdir(deps_dir):
            sys.path.append(deps_dir) 

        # Load data references
        from arjuna.engine.data.factory import DataReference, DataReferences
        self.__contextual_data_references = DataReferences()
        self.__indexed_data_references = DataReferences()

        for linked_project in self.__linked_projects:
            cdrs, idrs = DataReference.load_all(linked_project.ref_conf)
            self.__contextual_data_references.update(cdrs)
            self.__indexed_data_references.update(idrs)

        cdrs, idrs  = DataReference.load_all(self.ref_config)
        self.__contextual_data_references.update(cdrs)
        self.__indexed_data_references.update(idrs)

        # Load localization data
        from arjuna.engine.data.localizer import Localizer
        self.__localizer = Localizer.load_all(self.ref_config)

        from arjuna.tpi.parser.yaml import Yaml
        from arjuna.interact.gui.auto.finder.withx import WithX
        fpath = self.ref_config.value(ArjunaOption.CONF_WITHX_FILE)
        creation_context= f"WithX.yaml file at {fpath}"
        if os.path.isfile(fpath):        
            wyaml = Yaml.from_file(fpath, allow_any=True)
            if wyaml is not None:
                self.__common_withx_ref = WithX(wyaml.as_map())

        self.__start_bmproxy(self.ref_config)

        return self.ref_config

    @property
    def linked_projects(self):
        return self.__linked_projects

    @property
    def test_wise_container(self):
        return self.__current_test_wise_objects

    @property
    def withx_ref(self):
        return self.__common_withx_ref

    @property
    def contextual_data_references(self):
        return self.__contextual_data_references

    @property
    def indexed_data_references(self):
        return self.__indexed_data_references

    @property
    def localizer(self):
        return self.__localizer

    @property
    def logger(self):
        return self.__logger.arjuna_logger

    @property
    def data_store(self):
        return self.__data_store

    @property
    def test_session(self):
        return self.__test_session

    def get_config_value(self, query, *, cname=None):
        if cname is None:
            if type(query) is str:
                if query.find('.') != -1:
                    conf, _query = query.split(".", 1)
                    if self.has_config(conf):
                        cname = conf
                        query = _query
                    else:
                        cname = "ref"
                else:
                    cname = "ref"
            else:
                cname = "ref"

        return self.get_config(cname).value(query)

    @property
    def ref_config(self):
        return self.__thread_wise_ref_conf_map[threading.currentThread().name]


    def __get_thread_name(self):
        #multiprocessing.current_process().name
        return threading.current_thread().name

    def get_config(self, name):
        if name == "ref":
            current_proc_name = self.__get_thread_name()
            if current_proc_name == "MainThread":
                return self.ref_config
            else:
                return self.get_group_params()["config"]
        else:
            if not self.has_config(name):
                raise UndefinedConfigError(name, tuple(self.__config_map.keys()))
            return self.__config_map[name.lower()]

    def register_config(self, config):
        self.__config_map[config.name.lower()] = config

    def has_config(self, name):
        name = name.lower()
        return name in self.__config_map

    def get_group_params(self):
        return self.__thread_wise_group_params_map[self.__get_thread_name()]

    def register_group_params(self, **params):
        self.__thread_wise_group_params_map[self.__get_thread_name()] = params
        self.__thread_wise_ref_conf_map[self.__get_thread_name()] = params['config']

    def register_test_selector_for_group(self, selector):
        self.__thread_wise_test_selector_map[self.__get_thread_name()] = selector

    def register_pytest_command_for_group(self, group_command):
        self.__thread_wise_group_command_map[self.__get_thread_name()] = group_command

    def get_pytest_command_for_group(self):
        return self.__thread_wise_group_command_map[self.__get_thread_name()]

    def get_test_selector(self):
        try:
            return self.__thread_wise_test_selector_map[self.__get_thread_name()]
        except KeyError:
            raise TestSelectorNotFoundError()

    def register_test_meta_data(self, qual_name, test_meta_data):
        if qual_name not in self.__test_meta_data:
            self.__test_meta_data[qual_name] = test_meta_data

    def get_test_meta_data(self, qual_name):
        try:
            return self.__test_meta_data[qual_name]
        except KeyError as e:
            try:
                from arjuna import ArjunaOption
                return self.__test_meta_data[self.get_config("ref").value(ArjunaOption.PROJECT_NAME) + "." + qual_name]
            except KeyError:
                raise Exception("Test Meta data not found for test: {}. If you have a function with prefix 'check_', verify if it is decorated with @test. Existing entries: {}".format(qual_name, self.__test_meta_data.keys()))

class Arjuna:
    '''
        Facade of Arjuna framework.
        Contains static methods which wrapper an internal singleton class for easy access to top level Arjuna functions.
    '''

    ARJUNA_SINGLETON = None
    LOGGER = None
    __ARJ_COMMAND = None

    @classmethod
    def _set_command(cls, command):
        cls.__ARJ_COMMAND = command

    @classmethod
    def _get_command(cls):
        return cls.__ARJ_COMMAND

    @classmethod
    def init(cls, project_root_dir, cli_config=None, run_id=None, *, static_rid=False, linked_projects=[]):
        '''
            Returns reference test context which contains reference configuration.
            This reference test context merges central conf, project conf and central CLI options.
            Root directory is assumed as per the project structure.
            You can also provide an alternative root directory for test project.
        '''
        cls.ARJUNA_SINGLETON = ArjunaSingleton()
        if linked_projects is None:
            linked_projects = list()
        return cls.ARJUNA_SINGLETON.init(project_root_dir, cli_config, run_id, static_rid=static_rid, linked_projects=linked_projects)

    @classmethod
    def get_logger(cls):
        '''
            Returns framework logger.
        '''
        return cls.ARJUNA_SINGLETON.logger

    @classmethod
    def _get_allowed_log_contexts(cls):
        return cls.ARJUNA_SINGLETON.allowed_log_contexts

    @classmethod
    def _get_bmproxy_server(cls):
        return cls.ARJUNA_SINGLETON.bmproxy_server

    @classmethod
    def get_test_session(cls):
        '''
            Returns the current Test Session object.
        '''
        return cls.ARJUNA_SINGLETON.test_session

    @classmethod
    def register_config(cls, config):
        cls.ARJUNA_SINGLETON.register_config(config)

    @classmethod
    def has_config(cls, name):
        return cls.ARJUNA_SINGLETON.has_config(name)

    @classmethod
    def get_config(cls, name="ref"):
        '''
            Returns the configuration.
        '''
        return cls.ARJUNA_SINGLETON.get_config(name)

    @classmethod
    def get_configs(cls, *names):
        '''
            Returns the reference configuration.
        '''
        return [cls.get_config(name) for name in names]

    @classmethod
    def get_config_value(cls, query, *, cname=None):
        '''
            Returns the configuration value.
        '''
        return cls.ARJUNA_SINGLETON.get_config_value(query, cname=cname)

    @classmethod
    def get_data_ref(cls, name):
        try:
            return cls.ARJUNA_SINGLETON.contextual_data_references[name]
        except:
            try:
                return cls.ARJUNA_SINGLETON.indexed_data_references[name]
            except:
                raise Exception(f"No data reference found with name: {name}")

    @classmethod
    def get_dataref_value(cls, query="", *, bucket=None, context=None, index=None):
        '''
            Returns the data reference value for a given context.
        '''
        from arjuna.engine.data.reference import R
        return R(query, bucket=bucket, context=context, index=index)

    @classmethod
    def get_report_metadata(cls):
        return cls.ARJUNA_SINGLETON.test_wise_container

    @classmethod
    def get_gui_mgr(cls):
        '''
            Returns the central GUI Manager object that mangages namespaces.
        '''
        return cls.ARJUNA_SINGLETON.gui_mgr

    @classmethod
    def get_localizer(cls):
        return cls.ARJUNA_SINGLETON.localizer

    @classmethod
    def get_localized_str(cls, in_str, *, locale=None, bucket=None, strict=None):
        from arjuna.engine.data.localizer import L
        return L(in_str, locale=locale, bucket=bucket, strict=strict)

    @classmethod
    def get_data_store(cls):
        return cls.ARJUNA_SINGLETON.data_store

    @classmethod
    def get_withx_ref(cls):
        return cls.ARJUNA_SINGLETON.withx_ref

    @classmethod
    def get_group_params(cls):
        return cls.ARJUNA_SINGLETON.get_group_params()

    @classmethod
    def register_group_params(cls, **params):
        return cls.ARJUNA_SINGLETON.register_group_params(**params)

    @classmethod
    def register_test_selector_for_group(cls, selector):
        cls.ARJUNA_SINGLETON.register_test_selector_for_group(selector)

    @classmethod
    def get_test_selector(cls):
        return cls.ARJUNA_SINGLETON.get_test_selector()

    @classmethod
    def register_test_meta_data(cls, qual_name, test_meta_data):
        cls.ARJUNA_SINGLETON.register_test_meta_data(qual_name, test_meta_data)

    @classmethod
    def get_test_meta_data(cls, qual_name):
        return cls.ARJUNA_SINGLETON.get_test_meta_data(qual_name)

    @classmethod
    def register_pytest_command_for_group(cls, group_command):
        cls.ARJUNA_SINGLETON.register_pytest_command_for_group(group_command)

    @classmethod
    def get_pytest_command_for_group(cls):
        return cls.ARJUNA_SINGLETON.get_pytest_command_for_group()

    @classmethod
    def exit(cls):
        '''
            Clean-up and finalise resources currently opened by Arjuna.
        '''
        proxy_server = cls._get_bmproxy_server()
        if proxy_server is not None:
            proxy_server.stop()

    @classmethod
    def get_linked_projects(cls):
        return cls.ARJUNA_SINGLETON.linked_projects


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.core import checks
from arjuna.core.constant import BuiltInProp, RuleConditionType, RuleTargetType
from arjuna.core.error import InvalidSelectionRule

def none(value):
    if type(value) is str:
        return value.lower() == "none" and None or value
    else:
        return value is None and None or value

def custom_bool(value):
    if type(value) is bool:
        return value
    elif value.lower() in BOOL_MAP:
        return BOOL_MAP[value.lower()]
    else:
        raise InvalidSelectionRule(f"Provided unexpected value >>{value}<< for boolean context in rule. Allowed: [true/false/on/off/yes/no/0/1]")

built_in_prop_type = {
    BuiltInProp.PACKAGE : str,
    BuiltInProp.MODULE : str,
    BuiltInProp.NAME : str,
    BuiltInProp.QUAL_NAME : str,
    BuiltInProp.ID : str,
    BuiltInProp.PRIORITY : int,
    BuiltInProp.NAME : str,
    BuiltInProp.AUTHOR : str,
    BuiltInProp.IDEA : str,
    BuiltInProp.UNSTABLE : bool,
    BuiltInProp.COMPONENT : str,
    BuiltInProp.APP_VERSION : str,
    BuiltInProp.LEVEL : str,
    BuiltInProp.REVIEWED : bool,
}

BOOL_MAP = {
    'true': True,
    'false': False,
    'on': True,
    'off': False,
    'yes' : True,
    'no' : False
}

SYMBOLS_MAP = {
    'is': 'equal',
    'not': 'not_equal',
    'eq': 'equal',
    '=': 'equal',
    '==': 'equal',
    '!=': 'not_equal',
    'ne': 'not_equal',
    '~=': 'matches',
    '!~=': 'does_not_match',
    'matches': 'matches',
    '*=': 'partially_matches',
    '!*=': 'does_not_partially_match',
    'lt': 'less_than',
    '<': 'less_than',
    'le': 'less_or_equal',
    '<=': 'less_or_equal',
    'gt': 'greater_than',
    '>': 'greater_than',
    'ge': 'greater_or_equal',
    '>=': 'greater_or_equal',
}

def get_all_symbols():
    return SYMBOLS_MAP.keys()

__str_symbols = {
    'matches',
    '~=',
    '!~=',
    '\!~=',
    '*=',
    '!*=',
    '\!*=',
    'is',
    'not',
    'eq',
    'ne',
    '!=',
    '=',
    '==',
}

__number_symbols =  {
    'is',
    'not',
    'eq',
    'ne',
    '=',
    '==',
    '!=',
    'lt',
    '<',
    'le',
    '<=',
    'gt',
    '>',
    'ge',
    '>='
}

__appver_symbols = __str_symbols.union(__number_symbols)
__str_symbols = __str_symbols.union(__number_symbols)

__priority_symbols =  {
    'is' : 'is',
    'not': 'not',
    'eq': 'eq',
    'ne': 'ne',
    '=': '=',
    '==': '==',
    '!=': '!=',
    'lt': 'gt',
    '<': '>',
    'le': 'ge',
    '<=': '>=',
    'gt': 'lt',
    '>': '<',
    'ge': 'le',
    '>=': '<=',
}

# __str_rules = {
#     RuleConditionType.EQUAL, 
#     RuleConditionType.MATCHES, 
#     RuleConditionType.CONTAINS, 
#     RuleConditionType.PARTIALLY_MATCHES
# }

# __number_rules =  {
#     RuleConditionType.EQUAL,
#     RuleConditionType.NOT_EQUAL,
#     RuleConditionType.LESS_THAN,
#     RuleConditionType.LESS_OR_EQUAL,
#     RuleConditionType.GREATER_THAN,
#     RuleConditionType.GREATER_OR_EQUAL
# }

# __appver_rules = __str_rules.union(__number_rules)

# ALLOWED_CONDITIONS = {
#     str : __str_rules,
#     bool : {
#             RuleConditionType.EQUAL, 
#             RuleConditionType.NOT_EQUAL
#         },
#     int : __number_rules,
#     float : __number_rules
# }

ALLOWED_SYMBOLS = {
    str: __str_symbols,
    bool: ['is', 'not', '=', '==', 'eq', '!=', 'ne'],
    int: __number_symbols,
    float: __number_symbols
}

VALUE_CHECKERS = {
    RuleConditionType.EQUAL : checks.are_equal,
    RuleConditionType.NOT_EQUAL : checks.are_not_equal,
    RuleConditionType.LESS_THAN : checks.less_than,
    RuleConditionType.LESS_OR_EQUAL : checks.less_or_equal,
    RuleConditionType.GREATER_THAN : checks.greater_than,
    RuleConditionType.GREATER_OR_EQUAL : checks.greater_or_equal,
    RuleConditionType.MATCHES : checks.match_with_ignore_case,
    RuleConditionType.DOES_NOT_MATCH : checks.does_not_match_with_ignore_case,
    RuleConditionType.PARTIALLY_MATCHES : checks.partially_match_with_ignore_case,
    RuleConditionType.DOES_NOT_PARTIALLY_MATCH : checks.does_not_partially_match_with_ignore_case,
    RuleConditionType.CONTAINS : checks.contains,
    RuleConditionType.IS_SUBSET : checks.is_subset,
    RuleConditionType.HAS_INTERSECTION : checks.has_intersection,
    RuleConditionType.NO_INTERSECTION : checks.has_no_intersection,
}

CONVERTER_MAP = {
    int: int,
    str: str,
    float : float,
    bool: custom_bool
}

def _validate_symbol(symbol):
    if symbol not in SYMBOLS_MAP:
        raise InvalidSelectionRule("Invalid condition symbol >>{}<< used. Allowed: {}".format(symbol, tuple(SYMBOLS_MAP.keys())))

def _validate_allowed_symbol(target, target_type, symbol):
    if target == "app_version":
        if symbol not in __appver_symbols:
            raise InvalidSelectionRule("For app_version property, you have used an unexpected condition [{}]. Allowed: {}.".format(symbol, __appver_symbols))
    elif symbol not in ALLOWED_SYMBOLS[target_type]:
        raise InvalidSelectionRule("[{}] property is of type [{}]. You have used an unexpected condition [{}]. Allowed: {}.".format(target, target_type.__name__, symbol, ALLOWED_SYMBOLS[target_type]))

def convert_to_condition(target, target_type, symbol):
    symbol = symbol.strip().lower()
    _validate_symbol(symbol)
    _validate_allowed_symbol(target, target_type, symbol)
    if target != "priority":
        return RuleConditionType[SYMBOLS_MAP[symbol].upper()]
    else:
        return RuleConditionType[SYMBOLS_MAP[__priority_symbols[symbol]].upper()]


def get_value_checker_for_symbol(condition):
    return VALUE_CHECKERS[condition]

def convert_expression(target, target_type, expression):
    if expression.lower().strip() == 'none':
        return None
    if is_builtin_prop(target):
        return CONVERTER_MAP[target_type](expression)
    else:
        return expression

ALLOWED_TAG_CONTAINERS = {'tag', 'tags', 'bug', 'bugs', 'env', 'envs'}

TAG_CONTAINER_MAP = {
    'tag': 'tags', 
    'tags': 'tags', 
    'bug': 'bugs',  
    'bugs': 'bugs', 
    'env': 'envs', 
    'envs': 'envs' 
}

def get_tag_container(container):
    container = container.strip().lower()
    if container in ALLOWED_TAG_CONTAINERS:
        #raise InvalidSelectionRule("Unrecognized tag container [{}]. Allowed: {}".format(container, ALLOWED_TAG_CONTAINERS))
        return TAG_CONTAINER_MAP[container]
    else:
        return container

def validate_built_in_props(props):
    for k,v in props.items():
        if is_builtin_prop(k.upper()):
            expected_type = built_in_prop_type[BuiltInProp[k.upper()]]
            actual_type = type(v)
            if v is not None and actual_type is not expected_type:
                raise Exception("Built-in property {} should of type {}. Found {} of type {}".format(
                    k,
                    expected_type,
                    v,
                    actual_type
                ))

def get_value_type(built_in_prop_name):
    return built_in_prop_type[BuiltInProp[built_in_prop_name.upper()]]

def is_builtin_prop(name):
    return name.upper() in BuiltInProp.__members__

__type_defaults = {
    bool: False,
    str: None,
    int: None,
    float: None
}
def get_default_value_for_type(in_type, name):
    if name == "priority":
        return 5
    elif name == "app_version":
        return "0.0.0"
    else:
        return __type_defaults[in_type]



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Locatable:
    '''
        Represents an object that is locatable using **GuiWidgetMetaData** in a **Gui**.
    '''

    def __init__(self, gui, wmd):
        # super().__init__(automator.config) #, obj_name)
        self.__gui = gui
        self.__automator = gui._automator
        self.__wmd = wmd
        self.__located = False
        self.__located_with = None 

        # self.__parent = parent

    @property
    def gui(self):
        '''
            **Gui** object containing this **Locatable** object.
        '''
        return self.__gui
        
    @property
    def _automator(self):
        return self.__automator

    @property
    def _wmd(self):
        return self.__wmd

    @property
    def _is_located(self):
        return self.__located

    @property
    def located_with(self):
        '''
            Identifier which located this Locatable. Useful to know if multiple identifiers are present in GuiWidgetLocator.
        '''
        return self.__located_with

    @located_with.setter
    def _located_with(self, locator_tuple):
        self.__located = True
        self.__located_with = locator_tuple


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.value import Value

class CliArgsConfig:

    def __init__(self, arg_dict):
        self.__option_map = {
            "ao" : dict(),
            "uo" : dict()
        }

        def update_map(otype, arg_dict):
            if otype in arg_dict:
                opt_list = arg_dict[otype]
                if opt_list is not None:
                    self.__option_map[otype].update(dict(arg_dict[otype]))
                del arg_dict[otype]
            else:
                self.__option_map[otype] = dict()
        
        update_map("ao", arg_dict)
        update_map("uo", arg_dict)
        self.__option_map["ao"].update(arg_dict)
        arg_dict.clear()

    @property
    def arjuna_options(self):
        return self.__option_map["ao"]

    @property
    def user_options(self):
        return self.__option_map["uo"]

    def as_map(self):
        return {
            "arjuna_options": self.arjuna_options,
            "user_options": self.user_options,
        }



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oauthlib.oauth2 import BackendApplicationClient, MobileApplicationClient
from requests_oauthlib import OAuth2Session

from .session import HttpSession

from bs4 import BeautifulSoup
import urllib.parse

class OAuthSession(HttpSession):
    '''
        Base Class for all types of OAuth Http Sessions.

        Arguments:
            session: `requests_oauthlib` session object
            url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
    '''

    def __init__(self, *, session, url):
        super().__init__(url=url, _auto_session=False)
        self._set_session(session)
        self.__token = None

    @property
    def token(self):
        '''
            OAuth Token created by this session.
        '''
        return self.__token

    def _set_outh_token(self, token):
        self.__token = token
        self._session.headers.update({'Authorization': 'Bearer ' + self.token})

    def create_new_session(self, url, *, request_content_handler=None):
        '''
            Create a new HttpSession object. OAuth token of this session is attached to the new session.

            Arguments:
                url: Base URL for the new HTTP session.
                content-type: Default request content type for the new session.
        '''
        from .http import Http
        session = Http.session(url=url, oauth_token=self.token, request_content_handler=request_content_handler)
        session.add_cookies(self._session.cookies)
        return session


class OAuthClientGrantSession(OAuthSession):
    '''
        Creates token using OAuth's Resource Owner Client Credentials Grant Type.
        Uses BackendApplicationClient from requests_oauthlib.

        Keyword Arguments:
            url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
            client_id: Client ID
            client_secret: Client Secret
            token_url: Token URL
    '''

    def __init__(self, *, url, client_id, client_secret, token_url):
        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)
        super().__init__(session=oauth, url=url)
        token = oauth.fetch_token(token_url=token_url, 
                                  client_id=client_id,
                                  client_secret=client_secret)
        self._set_outh_token(token['access_token'])


class OAuthImplicitGrantSession(OAuthSession):
    '''
        Creates token using OAuth's Implicit Code Grant Type.
        Uses MobileApplicationClient from requests_oauthlib.

        Keyword Arguments:
            url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
            client_id: Client ID
            scope: Scope
            redirect_uri: Redirect URI
            auth_url: Authorization URL
            auth_handler: A callback function to handle custom authroization logic. It will be called by providing session object, authorization URL and auth_args.
            **auth_args: Arbitray key-value pairs to be passed as arguments to the auth_handler callback.

        Note:
            Some sample auth_handler signatures:

                .. code-block:: python

                    auth_handler_1(oauth_session, auth_url, **kwargs)
                    auth_handler_2(oauth_session, auth_url, some_arg=None, another_arg="some_def_value")

    '''
    def __init__(self, *, url, client_id, scope, redirect_uri=None, auth_url, auth_handler=None, **auth_args):
        oauth = OAuth2Session(
            client=MobileApplicationClient(client_id=client_id),
            scope=scope,
            redirect_uri=redirect_uri,
        )
        super().__init__(session=oauth, url=url)

        auth_url, state = oauth.authorization_url(auth_url)

        token = None
        if auth_handler is None:
            token = outh.token_from_fragment(callback_url)
        else:
            token = auth_handler(self, auth_url, **auth_args)
        self._set_outh_token(token)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import threading
import types

from arjuna.core.adv.wrappers import ClassPassThrough


# From blog post by Daniel Holden at http://theorangeduck.com/page/synchronized-python
def sync_function(func):
    # reentrant lock to supprt recursion
    func.__lock__ = threading.RLock()

    @functools.wraps(func)
    def synced_func(*args, **kws):
        with func.__lock__:
            return func(*args, **kws)

    return synced_func


def sync_class(klass):
    for key in klass.__dict__:
        val = klass.__dict__[key]
        if type(val) is types.FunctionType:
            decorator = sync_function
            setattr(klass, key, decorator(val))

    return klass


# From blog post by Daniel Holden at http://theorangeduck.com/page/synchronized-python
# To use the following you should create self.lock = threading.RLock() in __init__
# Per instance basis lock
def sync_method(lock_name):
    def decorator(method):
        @functools.wraps(method)
        def synced_method(self, *args, **kws):
            lock = getattr(self, lock_name)
            with lock:
                return method(self, *args, **kws)

        return synced_method

    return decorator



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .gui import Gui
from .content import GuiAppContent
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.tpi.tracker import track
from arjuna.tpi.guiauto.source.page import GuiPageSource

@track("debug")
class GuiPage(GuiAppContent):
    '''
        Represents a GUI Page. 
        
        It is an implementation of **GuiAppContent**.

        Args:
            *args: Any number of positional argumnts. These are passed to the **prepare()** method if defined in inherited class.

        Keyword Arguments:
            source_gui: (Mandatory) The **Gui** object from where this **GuiPage** is created. 
            label: Label for the this **GuiPage**. If not provided, the class name is used as the label.
            gns_dir: Relative Root Directory for GNS file(s) associated with this **GuiPage**. Default is **page** directory in associated **GuiApp** namespace. If provided, it is considered relative to the namespace directory of associated **GuiApp**.
            gns_file_name: Name of GNS file associated with this **GuiPage**. If not provided, default is **<label>.yaml**.
            kwargs: Arbitrary keyword arugments. These are passed to the **prepare()** method if defined in inherited class.
    '''


    def __init__(self, *args, source_gui: Gui=None, label: str=None, gns_dir: str=None, gns_file_name: str=None, **kwargs):
        # app = isinstance(source_gui, GuiApp) and source_gui or source_gui.app
        super().__init__(automator=source_gui._automator, label=label, gns_dir=gns_dir, gns_file_name=gns_file_name)
        self.app.ui = self
        from arjuna import ArjunaOption
        if self.config.value(ArjunaOption.BROWSER_NETWORK_RECORDER_AUTOMATIC):
            self._automator.network_recorder.current_title = self.label
        self._load(*args, **kwargs)

    @property
    def source(self) -> GuiPageSource:
        '''
           **GuiPageSource** object for this **Gui**.
        '''
        return self._automator.source



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto

class ArjunaOption(Enum):
    '''
        Represents all built-in configuration options for Arjuna.

        Any option name which is does not correspond to ArjunaOption enum constant is treated as a user defined option.
    '''
    
    ARJUNA_ROOT_DIR = auto()
    '''Root Directory of Arjuna Installed/Imported in a session'''

    ARJUNA_EXTERNAL_IMPORTS_DIR = auto()
    '''Directory of third party libs directly included in Arjuna.'''

    LOG_NAME = auto()
    '''Name of Arjuna's log file'''

    RUN_ID = auto()
    '''An alnum string representing current test run. Default is **mrun**'''

    RUN_SESSION_NAME = auto()
    '''Current session name.'''

    RUN_HOST_OS = auto()
    '''Host Operating System type: Windows/Mac/Linux.'''

    LOG_FILE_LEVEL = auto()
    '''Minimum level for a message to be logged to log file.'''

    LOG_CONSOLE_LEVEL = auto()
    '''Minimum level for a message to be displayed on console'''

    LOG_ALLOWED_CONTEXTS = auto()
    '''Allowed context strings for logging (file as well as display). Messages without contexts always get logged.'''

    L10N_LOCALE = auto()
    '''Default Locale type to be used for Localization call. Values as per arjuna.tpi.constant.Locale'''

    L10N_STRICT = auto()
    '''Sets Localization mode to strict. Default is False.'''

    L10N_DIR = auto()
    '''Directory containing Localization files.'''

    PROJECT_NAME = auto()
    '''Test Project Name'''

    PROJECT_ROOT_DIR = auto()
    '''Test Project Root Directory'''

    CONF_PROJECT_FILE = auto()
    '''Project conf file path.'''

    TESTS_DIR = auto()
    '''Directory containing test modules.'''

    HOOKS_DIR = auto()
    '''Arjuna Hooks directory.'''

    REPORTS_DIR = auto()
    '''Root directory for test reports.'''

    REPORT_FORMATS = auto()
    '''Formats for Report Generation. XML/HTML'''

    REPORT_DIR = auto()
    '''Reporting directory for current test run under REPORTS_DIR. Name is generated with RUN_ID and Current Timestamp. With --static-rid CLI switch, timestamp is not appended.'''

    REPORT_XML_DIR = auto()
    '''Directory containing report.xml for current test run.'''

    REPORT_HTML_DIR = auto()
    '''Directory containing report.html for current test run.'''

    REPORT_GROUP_RENAME = auto()
    '''If True, for run-group command, reports are created without session and stage prefixes.'''

    REPORT_SCREENSHOTS_ALWAYS = auto()
    '''If True, Screenshots are always show in Report, else they are not shown for passed tests. Default is False.'''

    REPORT_NETWORK_ALWAYS = auto()
    '''If True, Network packets are always show in Report, else they are not shown for passed tests. Default is False.'''

    REPORT_NETWORK_FILTER = auto()
    '''If True, in reporting, the request/response for static resources like image files, css etc will be excluded. Only HTML/JSON/XML content is included. Default is True'''

    LOG_DIR = auto()
    '''Directory containing arjuna.log for current test run.'''

    SCREENSHOTS_DIR = auto()
    '''Directory containing screenshots for current test run.'''

    TOOLS_DIR = auto()
    '''Directory containing external tool binaries in Arjuna test project.'''

    TOOLS_BMPROXY_DIR = auto()
    '''Directory containing BrowerMob Proxy binaries.'''

    DEPS_DIR = auto()
    '''Directory containing external tool binaries in Arjuna test project.'''

    DBAUTO_DIR = auto()
    '''Directory containing Database interaction automation related input files.'''

    DBAUTO_SQL_DIR = auto()
    '''Directory containing SQL files for Database interaction automation.'''

    TEMP_DIR = auto()
    '''Temporary directory for this session.'''

    CONF_DIR = auto()
    '''Test Project configuration directory'''

    CONF_DATA_FILE = auto()
    '''File that contains all data configurations.'''

    CONF_ENVS_FILE = auto()
    '''File that contains all environment configurations.'''

    CONF_SESSIONS_FILE = auto()
    '''File that contains all test session definitions.'''

    CONF_STAGES_FILE = auto()
    '''Directory that contains all test stage definitions.'''

    CONF_GROUPS_FILE = auto()
    '''Directory that contains all test group definitions.'''

    CONF_WITHX_FILE = auto()
    '''Path of withx.yaml file used for writing custom locators for Gui Automation.'''

    DATA_DIR = auto()
    '''Directory containing data files in test project.'''

    DATA_SRC_DIR = auto()
    '''Directory containing data source files in test project.'''

    DATA_REF_DIR = auto()
    '''Directory containing contextual data reference files in test project.'''

    DATA_REF_CONTEXTUAL_DIR = auto()
    '''Directory containing contextual data reference files in test project.'''

    DATA_REF_INDEXED_DIR = auto()
    '''Directory containing indexed data reference files in test project.'''

    DATA_FILE_DIR = auto()
    '''Directory containing files used as file data.'''

    APP_URL = auto()
    '''Base URL for a Web App. Used by launch() method if url is not specified for GuiApp.'''

    BROWSER_NAME = auto()
    '''Browser Name for Gui Automation. Chrome/Firefox. Default is Chrome'''

    BROWSER_HEADLESS = auto()
    '''Sets headless mode for browser for GUI Automation. Default is False.'''

    BROWSER_VERSION = auto()
    '''Browser Version for GUI Automation.'''

    BROWSER_MAXIMIZE = auto()
    '''Browser is maximized in GUI Automation after launch. Default is False.'''

    BROWSER_DIM_HEIGHT = auto()
    '''Browser Height for GUI Automation. If not set, Arjuna does not change the height of browser.'''

    BROWSER_DIM_WIDTH = auto()
    '''Browser Width for GUI Automation. If not set, Arjuna does not change the width of browser.'''

    BROWSER_BIN_PATH = auto()
    '''Path of the Browser binary on test system.'''

    BROWSER_NETWORK_RECORDER_ENABLED = auto()
    '''If True, Arjuna uses BrowserMob proxy, if available in test project, to capture Network requests made by browser. Default is False.'''

    BROWSER_NETWORK_RECORDER_AUTOMATIC = auto()
    '''If True, when a browser is launched, Arjuna automatically starts capturing all traffic. Default is False'''

    ALLOW_INSECURE_SSL_CERT = auto()
    '''If True, insecure SSL certificates are allowd. Default is True'''

    GUIAUTO_NAME = auto()
    '''Engine name. Currently set to Selenium which is the only supported engine.'''

    GUIAUTO_DIR = auto()
    '''Root directory of all Gui automation relation directories and files'''

    GUIAUTO_NAMESPACE_DIR = auto()
    '''Root directory of all Gui Namespace (GNS) files.'''

    GUIAUTO_DEF_MULTICONTEXT = auto()
    '''Sets multi context mode for GNS files. Currently not processed.'''

    GUIAUTO_CONTEXT = auto()
    '''Gui Automation Context. Currently not processed.'''

    SCROLL_PIXELS = auto()
    '''Number of pixels for each scroll call in Gui Automation. Default is 100.'''

    GUIAUTO_MAX_WAIT = auto()
    '''Maximum time for a Gui element locating or waitable interaction to occur. Uses Dynamic Wait. Expressed in seconds. Default is 60.'''

    GUIAUTO_SLOMO_ON = auto()
    '''Sets slow motion mode for Gui Automation. Default is False.'''

    GUIAUTO_SLOMO_INTERVAL = auto()
    '''Time Interval between successive Gui Automation actions when Slow Motion mode is ON. Expressed in seconds. Default is 2'''

    MOBILE_OS_NAME = auto()
    '''Mobile OS Name. iOs/Android. Default is Android.'''

    MOBILE_OS_VERSION = auto()
    '''Mobile OS Version. No default set.'''
    
    MOBILE_DEVICE_NAME = auto()
    '''Mobile Device name. No default set.'''

    MOBILE_DEVICE_UDID = auto()
    '''Mobile Device UDID. No default set.'''

    MOBILE_APP_FILE_PATH = auto()
    '''Mobile App path on test system. No default set.'''

    SELENIUM_DRIVER_PROP = auto()
    '''Selenium Environment variable for browser driver as per chosen browser. Automatically set as per chosen browser. Default is webdriver.chrome.driver'''

    SELENIUM_DRIVERS_DIR = auto()
    '''Root Directory containing OS specific browser drivers for Selenium. Has an impact only if SELENIUM_DRIVER_DOWNLOAD is set to False.'''

    SELENIUM_DRIVER_PATH = auto()
    '''Absolute path of Selenium browser driver. Automatically set to WebDriverManager's downloaded driver if SELENIUM_DRIVER_DOWNLOAD is True, else automatically set as per the test project structure, OS and browser.'''

    SELENIUM_DRIVER_DOWNLOAD = auto()
    '''Instructs Arjuna to automatically download Selenium browser driver for chosen browser. Default is True.'''

    SELENIUM_SERVICE_URL = auto()
    '''Selenium's Service URL. If set, Arjuna does not launch the browser service and uses this URL as the service URL.'''

    APPIUM_SERVICE_URL = auto()
    '''Appium Service URL. Currently not processed.'''

    APPIUM_AUTO_LAUNCH = auto()
    '''Instructs Arjuna to launch Appium programmatically. Default is True. Currently not processed.'''

    IMG_COMP_MIN_SCORE = auto()
    '''A fraction that represents minimum image comparison score to decide on an image match. Default is 0.7. Currently not processed.'''

class TimeUnit(Enum):
    '''
        Allowed time unit types.
    '''

    MILLI_SECONDS = auto()
    SECONDS = auto()
    MINUTES = auto()

class BrowserName(Enum):
    '''
        Allowed browser names for Gui Automation.
    '''

    CHROME = auto()
    FIREFOX = auto()

import locale
import re
__locales = [i.upper() for i in locale.locale_alias.keys() if re.match('^[\w_]+$', i)]

Locale = Enum('Locale', dict(zip(__locales, range(len(__locales)))))
Locale.__doc__ = '''Allowed locale names in Arjuna.'''


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from haralyzer import HarParser, HarPage
from arjuna.tpi.httpauto.request import HttpRequest
from arjuna.tpi.httpauto.request import HttpResponse

from arjuna.tpi.helper.arjtype import NetworkPacketInfo

class Har:

    def __init__(self, pydict):
        self.__har = pydict
        self.__har_parser = HarParser(pydict)

    def get_pagewise_network_packet_info(self, filter_resources=True):
        page_network_info = []
        from arjuna import Arjuna, ArjunaOption
        for har_page in self.__har_parser.pages:
            sub_network_packets = []
            if filter_resources:
                entries = har_page.filter_entries(content_type=".*(text/plain|application/xhtml+xml|text/html|json|application/xml|text/xml)")
            else:
                entries = har_page.entries
            for entry in entries:
                req_method = entry['request']['method']
                req_url = entry['request']['url']
                req_headers= {header['name']: header['value'] for header in entry['request']['headers']}
                if 'content' in entry['request']:
                    req_content = entry['request']['content']['text']
                else:
                    req_content = None

                req_repr = HttpRequest.repr_as_str(
                    method=req_method,
                    url=req_url,
                    headers=req_headers,
                    content=req_content
                )

                resp_status = entry['response']['status']
                resp_status_text = entry['response']['statusText']
                resp_headers = {header['name']: header['value'] for header in entry['response']['headers']}
                if 'content' in entry['response']:
                    if 'text' in entry['response']['content']:
                        resp_content = entry['response']['content']['text']
                    else:
                        resp_content = None
                else:
                    resp_content = None

                resp_repr = HttpResponse.repr_as_str(
                    status_code=resp_status,
                    status_msg=resp_status_text,
                    headers=resp_headers,
                    content=resp_content
                )

                sub_network_packets.append(
                    NetworkPacketInfo(
                        label=f"{req_method} {req_url}",
                        request=req_repr,
                        response=resp_repr,
                        sub_network_packets=[]
                    )
                )

            if sub_network_packets:
                page_network_info.append(
                    NetworkPacketInfo(
                        label=har_page.title,
                        request="",
                        response="",
                        sub_network_packets=sub_network_packets
                    )
                )

        return page_network_info

    @classmethod
    def from_har(cls, harobj):
        try:
            return Har(harobj)
        except:
            raise Exception("Invalid HAR object.")

    @classmethod
    def from_file(cls, file_path):
        with open(file_path, "r") as f:
            return Har(json.loads(f.read()))



import re
from arjuna.core.constant import GuiElementType

from ._with import GuiGenericLocator
from .enums import GenericLocateWith

class LocatorTranslator:
    '''
        Translates With Constructs to what underlying engines understand.
    '''

    BASIC_LOCATORS = {
        GenericLocateWith.ID,
        GenericLocateWith.NAME,
        GenericLocateWith.XPATH,
        GenericLocateWith.IMAGE,
        GenericLocateWith.INDEX,
        GenericLocateWith.WINDOW_TITLE,
        GenericLocateWith.WINDOW_PTITLE,
        GenericLocateWith.JS,
    }

    NEED_TRANSLATION = {
        GenericLocateWith.TAG : GenericLocateWith.TAG_NAME,
        GenericLocateWith.LINK : GenericLocateWith.PARTIAL_LINK_TEXT,
        GenericLocateWith.FLINK : GenericLocateWith.LINK_TEXT,
        GenericLocateWith.SELECTOR : GenericLocateWith.CSS_SELECTOR,
    }

    XTYPE_LOCATORS = {
        GuiElementType.TEXTBOX: "//input[@type='text']",
        GuiElementType.PASSWORD: "//input[@type='password']",
        GuiElementType.LINK: "//a",
        GuiElementType.BUTTON: "//input[@type='button']",
        GuiElementType.SUBMIT_BUTTON: "//input[@type='submit']",
        GuiElementType.DROPDOWN: "//select",
        GuiElementType.CHECKBOX: "//input[@type='checkbox']",
        GuiElementType.RADIO: "//input[@type='radio']",
        GuiElementType.IMAGE: "//img",
    }

    XPATH_LOCATORS = {
        GenericLocateWith.TEXT : "//*[contains(text(),'{}')]",
        GenericLocateWith.FTEXT : "//*[text()='{}']",
        GenericLocateWith.VALUE : "//*[@value='{}']",
        GenericLocateWith.TITLE : "//*[@title='{}']",
        GenericLocateWith.IMAGE_SRC : "//img[@src='{}']"
    }

    NAMED_ARG_LOCATORS = {
        GenericLocateWith.ATTR : "//*[contains(@{name},'{value}')]",
        GenericLocateWith.FATTR : "//*[@{name}='{value}']",
        GenericLocateWith.POINT : "return document.elementFromPoint({x}, {y})",
    }

    @classmethod
    def translate(cls, locator):
        from arjuna import log_debug
        rltype = locator.ltype
        rlvalue = locator.lvalue
        glvalue = None
        try:
            gltype = GenericLocateWith[rltype.upper()]
        except Exception as e:
            raise Exception("Invalid locator across all automators: {}={}. Error: {}".format(rltype, type(rlvalue), str(e)))
        else:
            log_debug("Processing locator: Type: {}; Value: {}".format(str(gltype), rlvalue))
            if gltype == GenericLocateWith.ELEMENT:
                glvalue = rlvalue
            elif gltype in cls.BASIC_LOCATORS:
                glvalue = rlvalue
            elif gltype in cls.NEED_TRANSLATION:
                glvalue = rlvalue
                gltype = cls.NEED_TRANSLATION[gltype]
            elif gltype in cls.XPATH_LOCATORS:
                glvalue = cls.XPATH_LOCATORS[gltype].format(rlvalue)
                gltype = GenericLocateWith.XPATH
            elif gltype == GenericLocateWith.POINT:
                glvalue = cls.NAMED_ARG_LOCATORS[gltype].format(**rlvalue)
            elif gltype in cls.NAMED_ARG_LOCATORS:
                glvalue = cls.NAMED_ARG_LOCATORS[gltype].format(**rlvalue)
                gltype = GenericLocateWith.XPATH
            elif gltype == GenericLocateWith.CLASSES:
                css_string = None
                if type(rlvalue) is str:
                    css_string = "." + rlvalue.replace('.', ' ').strip()
                else:
                    if type(rlvalue[0]) is str:
                        css_string = "." + rlvalue[0].replace('.', ' ').strip()
                    else:
                        css_string = "." + ".".join(rlvalue[0])
                glvalue = re.sub(r'\s+', '.', css_string)
                gltype = GenericLocateWith.CSS_SELECTOR
            else:
                raise Exception("Locator not supported yet by Arjuna: " + rltype)
    
        return GuiGenericLocator(gltype, glvalue)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid
import time
import logging
import shutil
from arjuna import ArjunaOption


from arjuna.tpi.config import Configuration

from arjuna.configure.configurator import TestConfigurator
from arjuna.drive.invoker.databroker import TestSessionDataBrokerHandler
from arjuna.interact.gui.gom.guimgr import GuiManager

class TestSessionController:
    
    def __init__(self):
        self.__id = uuid.uuid4()
        self.__DEF_CONF_NAME = "ref"
        self.__default_ref_config = None
        self.__config_map = {}
        self.__cli_central_config = None
        self.__cli_test_config = None
        self.__configurator = None
        self.__project_config_loaded = False
        self.__guimgr = None
        self.__session = None

    @property
    def id(self):
        return self.__id

    @property
    def configurator(self):
        return self.__configurator

    @property
    def data_broker(self):
        return self.__data_broker   

    @property
    def gui_manager(self):
        return self.__guimgr

    def init(self, root_dir, cli_config=None, run_id=None, linked_projects=[]):
        self.__configurator = TestConfigurator(root_dir, cli_config, run_id, linked_projects)
        ref_config = self.__configurator.ref_config
        data_env_confs = self.__configurator.file_confs
        self.__guimgr = GuiManager(ref_config)
        ref_conf = self._create_config(ref_config)
        self.__add_to_map(ref_conf)
        for run_env_conf in [self._create_config(econf, name=name) for name, econf in data_env_confs.items()]:
            self.__add_to_map(run_env_conf)

        return ref_conf

    def __msession_config(self, ref_conf_name):
        from arjuna import Arjuna
        if ref_conf_name is None:
            ref_conf_name = "ref"
        return Arjuna.get_config(ref_conf_name)

    def load_tests(self, *, dry_run=False, ref_conf_name=None, rules=None):
        from arjuna import Arjuna
        from arjuna.engine.session import MagicTestSession, YamlTestSession
        
        session_name = Arjuna.get_config().value(ArjunaOption.RUN_SESSION_NAME).lower()
        if session_name == "msession":
            ref_config = self.__msession_config(ref_conf_name)
            self.__session = MagicTestSession(ref_config, dry_run=dry_run, rules=rules)
        else:
            self.__session = YamlTestSession(session_name, ref_conf_name, dry_run=dry_run)

    def load_tests_for_stage(self, *, stage_name, dry_run=False, ref_conf_name=None):
        ref_config = self.__msession_config(ref_conf_name)
        from arjuna.engine.session import MagicTestSessionForStage
        self.__session = MagicTestSessionForStage(stage_name, ref_config, dry_run=dry_run)

    def load_tests_for_group(self, *, group_name, dry_run=False, ref_conf_name=None):
        ref_config = self.__msession_config(ref_conf_name)
        from arjuna.engine.session import MagicTestSessionForGroup
        self.__session = MagicTestSessionForGroup(group_name, ref_config, dry_run=dry_run)

    def run(self):
        self.__session.run()

    def _create_config(self, config, name=None):
        config = Configuration(
            self,
            name and name or self.__DEF_CONF_NAME,
            config
        )
        return config

    def finish(self):
        pass

    def __add_to_map(self, config):
        from arjuna import Arjuna
        Arjuna.register_config(config)

    def load_options_from_file(self, fpath, *, conf_stage):
        return self.configurator.load_options_from_file(fpath, conf_stage=conf_stage)

    def register_config(self, name, arjuna_options, user_options, *, conf_stage, parent_config=None):
        config = self.configurator.register_new_config(arjuna_options, user_options, parent_config=parent_config, conf_stage=conf_stage)
        conf = self._create_config(config, name=name)
        self.__add_to_map(conf)
        return conf

    def create_file_data_source(self, record_type, file_name, *arg_pairs):
        raise NotImplementedError()
        # response = self._send_request(
        #     ArjunaComponent.DATA_SOURCE,
        #     DataSourceActionType.CREATE_FILE_DATA_SOURCE,
        #     *arg_pairs
        # )
        # return response.get_data_source_id()

    def define_gui(self, automator, label=None, name=None, qual_name=None, def_file_path=None):
        return self.gui_manager.define_gui(automator, label=label, name=name, qual_name=qual_name, def_file_path=def_file_path)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

class CurrentTestWiseContainer:

    def __init__(self):
        self.__images = {}
        self.__network_packets = {}
        self.__test_node_id = None

    @property
    def current_test_node_id(self):
        return self.__test_node_id

    @current_test_node_id.setter
    def current_test_node_id(self, node_id):
        self.__test_node_id = node_id

    @property
    def images(self):
        try:
            return tuple(self.__images[threading.current_thread])
        except KeyError:
            return tuple()

    @property
    def network_packets(self):
        try:
            return tuple(self.__network_packets[threading.current_thread])
        except KeyError:
            return tuple()

    def _get_images_html(self):
        if not self.images: return ""
        html = "<p><p><h3>&nbsp;&nbsp;Screenshots and Images</h3><p>"
        if self.images:
            for image in self.images:
                fpath = "../screenshot/{}".format(image.file_name)
                #img_elem = '''<img src="data:image/png;base64,{}"/>'''.format(image.base64)
                #html += '<a href="{}" onclick="openModalImage(event)">{}</a>'.format(fpath, img_elem)

                html += '''<img class="screenshot" onclick="openModalImage(event)" href="{}" src="data:image/png;base64,{}"/>'''.format(fpath, image.base64)
        html += "<p><p>"
        return html

    def _get_packets_html(self):
        if not self.network_packets: return ""
        html = '<div class="requets"><h3>&nbsp;&nbsp;Network Packets</h3>'
        import html as py_html
        if self.network_packets:
            for np in self.network_packets:
                req_str = py_html.escape(str(np.request))
                res_str = py_html.escape(str(np.response))
                label = np.label
                if np.sub_network_packets:
                    expander = '<span class="expand">+</span>'
                else:
                    expander = ''
                if req_str != "":
                    html += f'<div class="reqcontent"><p><table class="noborder"><tr><td>{expander}</td><td><button data-request="{req_str}" data-response="{res_str}" onclick="openModal(event)">{label}</button></td></tr></p>'
                else:
                    html += f'<div class="reqcontent"><p><table class="noborder"><tr><td>{expander}</td><td>{label}</td></tr></p>'
                for redir in np.sub_network_packets:
                    req_str = py_html.escape(str(redir.request))
                    res_str = py_html.escape(str(redir.response))
                    label = redir.label
                    html += f'<div class="reqcontent"><p><tr class="redir"><td></td><td><button class="redir_button" data-request="{req_str}" data-response="{res_str}" onclick="openModal(event)">{label}</button></td></tr></p></div>'
                html += '</table></div>' # np
        html += '</div>' # headline
        html += '</div>' # id="image"
        return html

    def has_content(self):
        if self.images or self.network_packets:
            return True
        else:
            return False

    def as_report_html(self, *, include_images=True, include_network=True):
        if (not include_images) and (not include_network):
            return None
        html = '<div class="image">'
        if include_images:
            html += self._get_images_html()
        if include_network:
            html += self._get_packets_html()
        html += "</div>"
        return html

    def add_image(self, image):
        tname = threading.current_thread
        if tname not in self.__images:
            self.__images[tname] = []
        self.__images[tname].append(image)

    def add_network_packet_info(self, packet):
        tname = threading.current_thread
        if tname not in self.__network_packets:
            self.__network_packets[tname] = []
        self.__network_packets[tname].append(packet)


    def clear(self):
        tname = threading.current_thread
        if tname in self.__images:
            del self.__images[tname]
        if tname in self.__network_packets:
            del self.__network_packets[tname]




# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

def is_true(input):
    if type(input) is not bool:
        raise Exception("Not a boolean")
    return input is True


def is_false(input):
    if type(input) is not bool:
        raise Exception("Not a boolean")
    return input is False

def throw_exception_for_incompt_equal_same(context, actual, expected):
    if actual is None or expected is None:
        return

    if type(actual) != type(expected):
        msg = 'Checking for {} objects does not work for objects of different types.'.format(context)
        raise Exception(msg)

def are_equal(actual, expected):
    throw_exception_for_incompt_equal_same("equal/non-equal", actual, expected)
    return expected == actual

def are_not_equal(actual, expected):
    throw_exception_for_incompt_equal_same("equal/non-equal", actual, expected)
    return not are_equal(actual, expected)

def are_almost_equal(actual, expected, delta):
    diff = abs(expected - actual)
    delta2 = abs(delta)
    return diff <= delta2


def are_same(actual, expected):
    throw_exception_for_incompt_equal_same("same/different", actual, expected)
    return expected is actual


def are_different(actual, expected):
    throw_exception_for_incompt_equal_same("same/different", actual, expected)
    return expected is not actual

def match(target, pattern):
    if target is None: return False
    if re.match('^' + pattern + '$', target):
        return True
    else:
        return False

def partially_match(target, pattern):
    if target is None: return False
    if re.search(pattern, target):
        return True
    else:
        return False

def match_with_ignore_case(target, pattern):
    if target is None: return False
    if re.match('^' + pattern + '$', target, re.IGNORECASE):
        return True
    else:
        return False

def does_not_match_with_ignore_case(target, pattern):
    if target is None: return True
    if re.match('^' + pattern + '$', target, re.IGNORECASE):
        return False
    else:
        return True

def partially_match_with_ignore_case(target, pattern):
    if target is None: return False
    if re.search(pattern, target, re.IGNORECASE) is not None:
        return True
    else:
        return False

def does_not_partially_match_with_ignore_case(target, pattern):
    if target is None: return True
    if re.search(pattern, target, re.IGNORECASE) is not None:
        return False
    else:
        return True

are_not_same = are_different

def is_none(obj):
    return obj is None


def is_not_none(obj):
    return obj is not None


def contains(parent, child):
    return child in parent

def is_subset(left, right):
    return left.issubset(right)

def has_intersection(left, right):
    return bool(left.intersection(right))

def has_no_intersection(left, right):
    return bool(not left.intersection(right))

def does_not_contain(parent, child):
    return child not in parent

numbers = {int, float}

def __validate_args_as_numbers(func, left, right):
    if type(left) not in numbers or type(right) not in numbers:
        raise Exception("{} relation check is allowed only for numbers. Received >>{}<< of type: {} and >>{}<< of type {}.".format(
            func,
            left,
            type(left),
            right,
            type(right)
        ))


def less_than(left, right):
    return left < right

def less_or_equal(left, right):
    return left <= right

def greater_than(left, right):
    return left > right

def greater_or_equal(left, right):
    return left >= right



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arjuna.core.utils import data_utils
from arjuna.core.types import constants


class TextFile:
    '''
    Represents a text file in Read mode.

    Arguments:
        fpath: Absolute path of text file.
    '''

    def __init__(self, fpath):
        self._fpath = fpath
        self._f = open(self._fpath, "r")

    def close(self):
        '''
           Close the file handle.
        '''
        self._f.close()

    def read(self) -> str:
        '''
           Read contents of text file.

           Returns:
                Content of text file as a Python str object.
        '''
        return self._f.read()

    def _get_file(self):
        return self._f


class _TextResource(TextFile):
    def __init__(self, file_name):
        super().__init__(
        os.path.abspath(os.path.join(os.path.dirname(__file__), os.sep.join(["../..", "res", file_name]))))


class TextFileAsLines:
    '''
        Represents a text file in Read mode to get line by line content.

        Arguments:
            fpath: Absolute path of text file.

        Keyword Arguments:
            **formatters: Arbitrary key-value arguments to format a line before returning.

        Note:
            You can loop over this object:

                .. code-block:: python

                    for line in fobj:
                        # Do something about the content
                        print(line)
    '''

    def __init__(self, fpath, **formatters):
        self.__filereader = open(fpath, "r")
        self.__formatters = formatters

    def __iter__(self):
        return self

    def next(self) -> str:
        '''
            Get next line.

            Returns:
                Line as a Python str object.
        '''
        l = self.__filereader.readline()
        if not l:
            self.close()
            raise StopIteration()           
        elif l.strip().startswith('#'):
            return self.next()
        else:
            if self.__formatters:
                return l.rstrip().format(**self.__formatters)
            else:
                return l.rstrip()

    __next__ = next

    def close(self):
        '''
           Close the file handle.
        '''
        self.__filereader.close()


class DelimTextFileWithLineAsSeq:
    '''
        Represents a text file in Read mode to get line by line content split by defined delimiter.

        Arguments:
            fpath: Absolute path of text file.

        Keyword Arguments:
            delimiter: (Optional) Delimiter that separates different parts of line. Default is tab (\t)
            header_line_present: (Optional) If True, the first line is treated as a header line.

        Note:
            You can loop over this object:

                .. code-block:: python

                    for line_parts in fobj:
                        # Do something about the parts of a current line
                        print(line_parts)
    '''

    def __init__(self, fpath, *, delimiter="\t", header_line_present=True):
        self.__filereader = TextFileAsLines(fpath)
        self.delimiter = delimiter
        self.__headers = None
        self._populate_headers()

    def __iter__(self):
        return self

    def _populate_headers(self):
        try:
            self.__headers = self.next()
        except:
            raise Exception("Invalid input file data. Empty headers line.")

    def next(self) -> tuple:
        '''
            Get next line parsed based on delimiter as a tuple.

            Returns:
                Line as a Python tuple object.
        '''
        line = self.__filereader.next()
        return data_utils.split(line.rstrip(), self.delimiter)

    __next__ = next

    @property
    def headers(self) -> tuple:
        '''
            Get header line parsed based on delimiter as a tuple.

            Returns:
                Header line as a Python tuple object. None if header line is not present.
        '''
        return self.__headers

    def close(self):
        '''
           Close the file handle.
        '''
        self.__filereader.close()


class DelimTextFileWithLineAsMap:
    '''
        Represents a text file in Read mode to get line by line content split by defined delimiter.

        The first line is treated as a header line.

        Arguments:
            fpath: Absolute path of text file.

        Keyword Arguments:
            delimiter: (Optional) Delimiter that separates different parts of line. Default is tab (\t)

        Note:
            You can loop over this object:

                .. code-block:: python

                    for line_parts in fobj:
                        # Do something about the parts of a current line
                        print(line_parts)
    '''

    def __init__(self, fpath, *, delimiter="\t"):
        self.__filereader = DelimTextFileWithLineAsSeq(fpath, delimiter=delimiter)

    def __iter__(self):
        return self

    @property
    def headers(self) -> tuple:
        '''
            Get header line parsed based on delimiter as a tuple.

            Returns:
                Header line as a Python tuple object.
        '''
        return self.__filereader.headers

    def close(self):
        '''
           Close the file handle.
        '''
        self.__filereader.close()

    def next(self) -> dict:
        '''
            Get next line parsed based on delimiter converted into a Python dictionary based on header line.

            Returns:
                Line as a Python dict object.
        '''
        line_parts = self.__filereader.next()

        if len(self.headers) != len(line_parts):
            raise Exception(
                "Invalid input file data. Number of headers and data values do not match.{0}Headers:{1}{0}Data values:{2}{0}".format(
                    os.linesep, self.headers, line_parts))
        else:
            return dict(zip(self.headers, line_parts))

    __next__ = next


class Text:
    '''
    Provides factory methods for dealing with reading text file content in various forms.
    '''

    @classmethod
    def file_content(cls, fpath) -> str:
        '''
            Get content of a text file.

            Arguments:
                fpath: Absolute path of text file.

            Returns:
                Content of text file as Python str object.
        '''
        f = TextFile(fpath)
        content = f.read()
        f.close()
        return content

    @classmethod
    def file_lines(cls, fpath) -> TextFileAsLines:
        '''
            Get a reader for text file in Read mode to get line by line content.

            Arguments:
                fpath: Absolute path of text file.

            Returns:
                Arjuna's TextFileAsLines object.

            Note:
                You must explicitly close the returned object by calling its **close()** method.
        '''
        return TextFileAsLines(fpath)

    @classmethod
    def delimited_file(cls, fpath, *, delimiter="\t", header_line_present=True) -> 'DelimTextFileWithLineAsMap OR DelimTextFileWithLineAsSeq':
        '''
            Represents a text file in Read mode to get line by line content split by defined delimiter.

            Arguments:
                fpath: Absolute path of text file.

            Keyword Arguments:
                delimiter: (Optional) Delimiter that separates different parts of line. Default is tab (\t)
                header_line_present: (Optional) If True, the DelimTextFileWithLineAsMap is created, else DelimTextFileWithLineAsSeq is created to represent the file.

            Note:
                You can loop over the returned object:

                    .. code-block:: python

                        for line in fobj:
                            # Do something about the parts of a current line
                            print(line)

                The line is a tuple of parts for DelimTextFileWithLineAsSeq object and is a dictionary for DelimTextFileWithLineAsMap object.

            Note:
                You must explicitly close the returned object by calling its **close()** method.
        '''
        if header_line_present:
            return DelimTextFileWithLineAsMap(fpath, delimiter=delimiter)
        else:
            return DelimTextFileWithLineAsSeq(fpath, delimiter=delimiter)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.core.constant import ConfigStage

class Configurator:

    def __init__(self):
        self.__arjuna = None

    @property
    def _arjuna(self):
        return self.__arjuna

    @_arjuna.setter
    def _arjuna(self, singleton):
        self.__arjuna = singleton

    def builder(self, *, base_config: str="ref") -> 'ConfigBuilder':
        '''
            Creates a configuration builder object which takes the provided configuration as its basis.

            Keyword Arguments:
                base_config: The base `Configuration` object. All its options become a part of the new configuration. Its options can be overriden and new options can be added using the builder to create a new **Configuration** object. Default is reference configuration in CLI.

            Returns:
                **new** `ConfigBuilder` object
        '''
        from arjuna import Arjuna
        from arjuna.tpi.config import ConfigBuilder
        return ConfigBuilder(base_config=Arjuna.get_config(name=base_config), auto_name_gen=False, _conf_stage=ConfigStage.REFERENCE)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import io
import os

from urllib.parse import urlencode
from requests.auth import *
from requests_toolbelt.multipart.encoder import MultipartEncoder
from .session import HttpSession
from .response import HttpResponse
from .oauth import OAuthImplicitGrantSession, OAuthClientGrantSession

from collections import namedtuple
from arjuna.tpi.data.entity import data_entity

_HttpContent = namedtuple("_HttpContent", "content type")
_HttpField = namedtuple("_HttpField", "name value is_file content_type headers")

class Http:
    '''
    The Facade class for HTTP Automation.
    '''

    @classmethod
    def session(cls, *, url=None, oauth_token=None, request_content_handler=None, headers=None, max_redirects=None, auth=None, proxy=None):
        '''
            Create an HTTP Session. Does automatic cookie management.

            Keyword Arguments:
                url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
                oauth_token: OAuth 2.0 Bearer token for this session.
                request_content_handler: Default content type handler for requests sent in this session. Overridable in individual sender methods. Default is Http.content.urlencoded.
                headers: HTTP headers to be added to request headers made by this session.
                max_redirects: Maximum number of redirects allowed for a request. Default is 30.
                auth: HTTP Authentication object: Basic/Digest.
                proxy: Proxies dict to be associated with this session.
        '''
        return HttpSession(url=url, oauth_token=oauth_token, request_content_handler=request_content_handler, headers=headers, max_redirects=max_redirects, auth=auth, proxy=proxy, _auto_session=True)

    @classmethod
    def get(cls, route, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP GET request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().get(route, label=label, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def head(cls, route, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP HEAD request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().head(route, label=label, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def delete(cls, route, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP DELETE request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().delete(route, label=label, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def post(cls, route, *, content, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP POST request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().post(route, label=label, content=content, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def put(cls, route, *, content, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().put(route, label=label, content=content, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def patch(cls, route, *, content, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().patch(route, label=label, content=content, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def options(cls, route, *, content, label=None, xcodes=None, strict=False, headers=None, cookies=None, allow_redirects=True, auth=None, timeout: float=None, pretty_url=False, query_params=None, **named_query_params) -> HttpResponse:
        '''
        Sends an HTTP PUT request.

        Arguments:
            route: Absolute or relative URL. If relative, then `url` of this session object is pre-fixed.

        Keyword Arguments:
            label: Label for this request. If available, it is used in reports and logs.
            content: Content to be sent in this HTTP request. If passed as string, then content-type set in session is used using the content request handler. It can also be a dictionary with keys - 'content' and 'type'.
            xcodes: Expected HTTP response code(s).
            strict: If True in case of unexpected status code, an AssertionError is raised, else HttpUnexpectedStatusCodeError is raised.
            headers: Mapping of additional HTTP headers to be sent with this request.
            cookies: Python dict of cookies to send with request.
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            auth: HTTP Authentication object: Basic/Digest.
            timeout: How long to wait for the server to send data before giving up.
            pretty_url: If True, the query params are formatted using pretty URL format instead of usual query string which is the default.
            query_params: A mapping of key-values to be included in query string.
            **named_query_params: Arbitrary key/value pairs. These are appended to the query string of URL for this request.

        Note:
            **query_params** and **named_query_params** have the same goal.
            In case of duplicates, named_query_params override query_params.
        '''
        return HttpSession().options(route, label=label, content=content, xcodes=xcodes, strict=strict, headers=headers, cookies=cookies, allow_redirects=allow_redirects, auth=auth, timeout=timeout, pretty_url=pretty_url, query_params=query_params, **named_query_params)

    @classmethod
    def proxy(cls, url):
        '''
        Create a proxy dict.
        '''
        return {'http': url, 'https': url}

    class auth:
        '''
        HTTP Authentication Builder
        '''

        @classmethod
        def basic(cls, *, user, pwd):
            '''
            Create an HTTP Basic Authentication object.
            '''
            return HTTPBasicAuth(user, pwd)

        @classmethod
        def digest(cls, *, user, pwd):
            '''
            Create an HTTP Basic Authentication object.
            '''
            return HTTPDigestAuth(user, pwd)

    class content:

        @classmethod
        def get_content_type(cls, handler_method):
            mapping = {
                cls.blank: "text/html",
                cls.html: "text/html",
                cls.text: "text/html",
                cls.utf8: "text/html",
                cls.bytes: "text/html",
                cls.urlencoded: "application/x-www-form-urlencoded",
                cls.json: "application/json",
                cls.xml: "application/xml"
            }
            return mapping[handler_method]

        @classmethod
        def blank(cls, content=""):
            '''
            Send empty content. Content-Type is sent as “text/html”
            '''
            return _HttpContent(content="", type=cls.get_content_type(cls.blank))

        @classmethod
        def html(cls, content=""):
            '''
            Send HTML content. Content-Type is sent as “text/html”
            '''
            return _HttpContent(content=content, type=cls.get_content_type(cls.html))

        text = html

        @classmethod
        def bytes(cls, content=""):
            '''
            Send bytes string. Content-Type is sent as “text/html”
            '''
            if content:
                content = io.BytesIO(content)
            return _HttpContent(content=content, type=cls.get_content_type(cls.bytes))

        @classmethod
        def utf8(cls, content=""):
            '''
            Send UTF-8 string. Content-Type is sent as “text/html”
            '''
            if content:
                content = content.encode("utf-8")
            return _HttpContent(content=content, type=cls.get_content_type(cls.utf8))

        @classmethod
        def urlencoded(cls, content=""):
            '''
            Send a dictionary of key-values in URL encoded format. Content-Type is sent as “application/x-www-form-urlencoded”
            '''
            if content:
                content = urlencode(content)
            return _HttpContent(content=content, type=cls.get_content_type(cls.urlencoded))

        @classmethod
        def json(cls, content=""):
            '''
            Send a dictionary of key-values as JSON. Content-Type is sent as “application/json”
            '''
            from json import JSONEncoder
            from arjuna.tpi.parser.yaml import YamlDict, YamlList
            from arjuna.tpi.parser.json import JsonDict, JsonList
            class _CustomEncoder(JSONEncoder):
                def default(self, o):
                    if isinstance(o, YamlDict) or isinstance(o, YamlList) or isinstance(o, JsonList) or isinstance(o, JsonDict):
                        return o.raw_object
                    return JSONEncoder.default(self, o)

            if type(content) is tuple:
                content = list(content)

            if content:
                if isinstance(content, JsonList) or isinstance(content, JsonDict):
                    content = content.raw_object
                content = json.dumps(content, cls=_CustomEncoder, indent=2)
            else:
                content = str(content)
            return _HttpContent(content=content, type=cls.get_content_type(cls.json))

        @classmethod
        def xml(cls, content=""):
            '''
            Send an XML string as XML. Content-Type is sent as “application/xml
            '''
            return _HttpContent(content=content, type=cls.get_content_type(cls.xml))

        @classmethod
        def file(cls, field_name, file_name, *, content_type='text/plain', headers=None):
            '''
            Upload a file and send as multipart data. Content-Type is sent as the content type got from multipart encoding.
            '''
            from arjuna import C, ArjunaOption
            file_path = os.path.join(C(ArjunaOption.DATA_FILE_DIR), file_name)
            encoder = MultipartEncoder(
                {field_name: (file_name, open(file_path, 'rb'), content_type, headers)}
            )
            return _HttpContent(content=encoder.to_string(), type=encoder.content_type)

        @classmethod
        def multipart(cls, *fields):
            '''
            Send the provided HTTP fields as multipart data. Content-Type is sent as the content type got from multipart encoding.
            '''
            from arjuna import C, ArjunaOption
            elist = list()
            for field in fields:
                if type(field) is dict:
                    for k,v in field.items():
                        elist.append((k, str(v)))
                elif isinstance(field, _HttpField):
                    if field.is_file:
                        file_path = os.path.join(C(ArjunaOption.DATA_FILE_DIR), field.value)
                        elist.append((field.name, (field.value, open(file_path, 'rb'), field.content_type, field.headers)))
                    else:
                        elist.append((field.name, str(field.value)))
            encoder = MultipartEncoder(elist)
            return _HttpContent(content=encoder.to_string(), type=encoder.content_type)

        @classmethod
        def custom(self, content, *, type):
            '''
            Send content with a custom content type.
            '''
            return _HttpContent(content=content, type=type)

    @classmethod
    def field(cls, name, value, is_file=False, content_type="text/plain", headers=None):
        return _HttpField(name=name, value=value, is_file=is_file, content_type=content_type, headers=headers)

    @classmethod
    def oauth_client_grant_session(cls, *, url, client_id, client_secret, token_url):
        '''
            Create OAuthClientGrantSession object.

            Creates token using OAuth's Resource Owner Client Credentials Grant Type.
            Uses BackendApplicationClient from requests_oauthlib.

            Keyword Arguments:
                url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
                client_id: Client ID
                client_secret: Client Secret
                token_url: Token URL
        '''
        return OAuthClientGrantSession(url=url, client_id=client_id, client_secret=client_secret, token_url=token_url)


    @classmethod
    def oauth_implicit_grant_session(cls, *, url, client_id, scope, redirect_uri=None, auth_url, auth_handler=None, **auth_args):
        '''
            Create OAuthImplicitGrantSession object.

            Creates token using OAuth's Implicit Code Grant Type.
            Uses MobileApplicationClient from requests_oauthlib.

            Keyword Arguments:
                url: Base URL for this HTTP session. If relative path is used as a route in sender methods like `.get`, then this URL is prefixed to their provided routes.
                client_id: Client ID
                scope: Scope
                redirect_uri: Redirect URI
                auth_url: Authorization URL
                auth_handler: A callback function to handle custom authroization logic. It will be called by providing session object, authorization URL and auth_args.
                **auth_args: Arbitray key-value pairs to be passed as arguments to the auth_handler callback.

            Note:
                Some sample auth_handler signatures:

                    .. code-block:: python

                        auth_handler_1(oauth_session, auth_url, **kwargs)
                        auth_handler_2(oauth_session, auth_url, some_arg=None, another_arg="some_def_value")

        '''
        return OAuthImplicitGrantSession(url=url, client_id=client_id, scope=scope, redirect_uri=redirect_uri, auth_url=auth_url, auth_handler=auth_handler, **auth_args)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess


def execute_windows_process(proc_name, arg_list=[]):
    '''
    Executes a windows process using the command shell in termination mode.

    Arguments:
    proc_name: Name of process
    arg_list: (Optional) A list of strings that are to be passed as command line arguments to said process.
    '''
    p = subprocess.Popen(
        ["cmd.exe", "/C", proc_name] + arg_list,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    stdout, stderr = p.communicate()

    return p.returncode, stdout, stderr


def kill_process(process_name):
    returncode, stdout, stderr = execute_windows_process("taskkill", ["/F", "/IM", process_name])
    if returncode != 0:
        raise Exception("Not able to kill process: " + process_name)


def start_service(name):
    returncode, stdout, stderr = execute_windows_process("sc", ["start", '"%s"'.format(name)])
    if returncode != 0 or not is_service_running(name):
        raise Exception("Unable to start service: " + name)


def stop_service(name):
    returncode, stdout, stderr = execute_windows_process("sc", ["stop", '"%s"'.format(name)])
    if returncode != 0 or is_service_running(name):
        raise Exception("Unable to stop service: " + name)


def is_service_running(name):
    returncode, stdout, stderr = execute_windows_process("sc", ["query", '"%s"'.format(name)])
    if returncode != 0:
        raise Exception("Not able to query service: " + name)
    else:
        return "RUNNING" in stdout



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from .source import *
from .reference import *
from arjuna.core.utils import file_utils
from arjuna.core.constant import DataRefType

def get_data_file_path(data_dir, fpath):
    if file_utils.is_absolute_path(fpath):
        if not file_utils.is_file(fpath):
            if file_utils.is_dir(fpath):
                raise Exception("Not a file: {}".format(fpath))
            else:
                raise Exception("File does not exist: {}".format(fpath))
        return fpath
    else:
        fpath = os.path.abspath(os.path.join(data_dir, fpath))
        if not file_utils.is_file(fpath):
            if file_utils.is_dir(fpath):
                raise Exception("Not a file: {}".format(fpath))
            else:
                raise Exception("File does not exist: {}".format(fpath))
        return fpath

def create_file_data_source(file_path, *, context, delimiter="\t"):
    from arjuna import Arjuna, ArjunaOption
    data_dir = Arjuna.get_config().value(ArjunaOption.DATA_SRC_DIR)
    file_path = get_data_file_path(data_dir, file_path)
    ds = None
    ext = file_path.lower()
    if ext.endswith(".csv") or ext.endswith(".txt"):
        ds = DsvFileMapDataSource(file_path, delimiter, context=context)
    elif ext.endswith(".xls"):
        ds = ExcelFileMapDataSource(file_path, context=context)
    elif ext.endswith(".ini"):
        ds = IniFileDataSource(file_path, context=context)
    else:
        raise Exception("This is not a default file format supported as a data source: " + file_path)
    return ds

from arjuna.tpi.helper.arjtype import CIStringDict

class DataReferences:

    def __init__(self):
        vars(self)['_store'] = CIStringDict()

    def as_dict(self):
        return vars(self)['_store'].orig_dict

    def __getitem__(self, name):
        return self._store[name]

    def __setitem__(self, name, value):
        self._store[name] = value

    def __getattr__(self, name):
        if type(name) is str and not name.startswith("__"):
            return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __str__(self):
        return str(self._store)

    def update(self, data_refs):
        vars(self)['_store'].update(data_refs.as_dict())


class DataReference:

    @classmethod
    def load_all(cls, ref_config):
        from arjuna.tpi.constant import ArjunaOption
        crefs = DataReferences()
        contextual_data_ref_dir = ref_config.value(ArjunaOption.DATA_REF_CONTEXTUAL_DIR)
        if os.path.isdir(contextual_data_ref_dir):
            for fname in os.listdir(contextual_data_ref_dir):
                if not (fname.lower().endswith("xls") or fname.lower().endswith("yaml")) :
                    continue
                crefs[os.path.splitext(fname)[0]] = cls.create_contextual_data_ref(ref_config, fname)

        irefs = DataReferences()
        indexed_data_ref_dir = ref_config.value(ArjunaOption.DATA_REF_INDEXED_DIR)
        if os.path.isdir(indexed_data_ref_dir):
            for fname in os.listdir(indexed_data_ref_dir):
                if not (fname.lower().endswith("xls") or fname.lower().endswith("yaml")) :
                    continue
                irefs[os.path.splitext(fname)[0]] = cls.create_indexed_data_ref(ref_config, fname)
        return crefs, irefs

    @classmethod
    def __create_file_data_ref(cls, ref_config, file_path, type):
        ext = file_path.lower()
        if not ext.endswith("xls") and not ext.endswith("yaml"):
            raise Exception("Unsupported file extension for data reference: {}. Allowed: [xls, yaml]".format(file_path))

        from arjuna import Arjuna, ArjunaOption
        if type == DataRefType.CONTEXTUAL:
            data_dir = ref_config.value(ArjunaOption.DATA_REF_CONTEXTUAL_DIR)
            file_path = get_data_file_path(data_dir, file_path)
            if file_path.lower().endswith("xls"):
                return ExcelContextualDataReference(file_path)
            else:
                return YamlContextualDataReference(file_path)
        elif type == DataRefType.INDEXED:
            data_dir = ref_config.value(ArjunaOption.DATA_REF_INDEXED_DIR)
            file_path = get_data_file_path(data_dir, file_path)
            if file_path.lower().endswith("xls"):
                return ExcelIndexedDataReference(file_path)
            else:
                return YamlIndexedDataReference(file_path)

    @classmethod
    def create_contextual_data_ref(cls, ref_config, file_path):
        return cls.__create_file_data_ref(ref_config, file_path, DataRefType.CONTEXTUAL)

    @classmethod
    def create_indexed_data_ref(cls, ref_config, file_path):
        return cls.__create_file_data_ref(ref_config, file_path, DataRefType.INDEXED)

